# -*- coding: utf-8 -*-
#
# Zeek documentation build configuration file, created by sphinx-quickstart
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os

extensions = []

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('ext'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions += ['zeek', 'sphinx.ext.todo', 'zeek_pygments', 'literal-emph']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Zeek'
copyright = u'2019-2021, The Zeek Project'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
#

version = u"source"

try:
    # Use the actual Zeek version if available
    with open('../VERSION', 'r') as f:
        version = f.readline().strip()
except:
    try:
        import git
        import re

        repo = git.Repo(os.path.abspath('.'))
        version = u"git/master"

        version_tag_re = 'v\d+\.\d+(\.\d+)?'
        version_tags = [t for t in repo.tags if
                t.commit == repo.head.commit and
                re.match(version_tag_re, str(t))
                ]
        # Note: sorting by tag date doesn't necessarily give correct
        # order in terms of version numbers, but doubtful that will ever be
        # a problem (if we ever do re-tag an old version number on a given
        # commit such that it is incorrectly found as the most recent version,
        # we can just re-tag all the other version numbers on that same commit)
        version_tags = sorted(version_tags, key=lambda t: t.tag.tagged_date)

        if version_tags:
            version = str(version_tags[-1])

    except:
        pass

# The full version, including alpha/beta/rc tags.
release = version

# In terms of the actual hyperlink URL, a more ideal/stable way to reference
# source code on GitHub would be by commit hash, but that can be tricky to
# update in a way that produces stable Sphinx/reST configuration: don't want
# to update the commit-hash for every Zeek commit unless it actually produces
# new content, and also don't want to accidentally make it easy for people to
# insert unreachable commits when manually running
# `zeek/ci/update-zeekygen-docs.sh`.
#
# We only have a few versions of docs that actually matter: `master` and
# `release/.*`, and the tip of those branches will always be in sync with
# auto-generated content by simply having `zeek/ci/update-zeekygen-docs.sh`
# change this to `release/.*` when needed.
zeek_code_version = 'release/4.2'
zeek_code_url = f'https://github.com/zeek/zeek/blob/{zeek_code_version}'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [".#*"]

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

highlight_language = 'none'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:
    # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_last_updated_fmt = '%B %d, %Y'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'analytics_id': 'UA-144186885-1',
    'collapse_navigation': False,
    'display_version': True,
    "style_external_links": True,
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> Documentation".
html_title = f'Book of Zeek ({release})'

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'images/zeek-logo-sidebar.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'images/zeek-favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

def setup(app):
    app.add_css_file("theme_overrides.css")
    from sphinx.highlighting import lexers
    from zeek_pygments import ZeekLexer
    lexers['zeek'] = ZeekLexer()
    app.add_config_value('zeek-code-url', zeek_code_url, 'env')

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {
#'**': ['localtoc.html', 'sourcelink.html', 'searchbox.html'],
#}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'zeek-docs'

# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'Zeek.tex', u'Zeek Documentation',
   u'The Zeek Project', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'zeek', u'Zeek Documentation',
     [u'The Zeek Project'], 1)
]

# -- Options for todo plugin --------------------------------------------
todo_include_todos=True



from __future__ import print_function

from . import zeek

import traceback
import sys

class Event(zeek.Event):

    warnings_emitted = set()

    def __new__(cls, *args, **kwargs):
        stack_info = traceback.extract_stack()[0]
        usage_file = stack_info[0]
        usage_line = stack_info[1]
        usage_text = stack_info[3]

        if (usage_file, usage_line) not in Event.warnings_emitted:
            print('File "{}", line {}: deprecated bro.event usage,'
                  ' use zeek.Event instead:\n    {}'.format(
                      usage_file, usage_line, usage_text),
                  file=sys.stderr)

            Event.warnings_emitted.add((usage_file, usage_line))

        return super(Event, cls).__new__(cls, *args, **kwargs)



# Zeekctl test plugin that defines a string to write to zeekctl-config.zeek
# using the old legacy BroControl API.

import BroControl.plugin

class BroctlConfigTest(BroControl.plugin.Plugin):
    def __init__(self):
        super(BroctlConfigTest, self).__init__(apiversion=1)

    def name(self):
        return "broctlconfigtest"

    def pluginVersion(self):
        return 1

    def init(self):
        return True

    def broctl_config(self):
        return 'redef TestVar = "this is a test";\nredef Test="another test";'



from pygments.lexer import RegexLexer, bygroups, include, words, bygroups
from pygments.token import *

def setup(Sphinx):
    pass

class ZeekLexer(RegexLexer):
    """
    For `Zeek <https://www.zeek.org/>`_ scripts.

    .. versionadded:: 2.5
    """
    name = 'Zeek'
    aliases = ['zeek']
    filenames = ['*.zeek']

    _hex = r'[0-9a-fA-F]'
    _float = r'((\d*\.?\d+)|(\d+\.?\d*))([eE][-+]?\d+)?'
    _h = r'[A-Za-z0-9][-A-Za-z0-9]*'

    tokens = {
        'root': [
            include('whitespace'),
            include('comments'),
            include('directives'),
            include('attributes'),
            include('types'),
            include('keywords'),
            include('literals'),
            include('operators'),
            include('punctuation'),

            (r'\b((?:[A-Za-z_][A-Za-z_0-9]*)(?:::(?:[A-Za-z_][A-Za-z_0-9]*))*)(?=\s*\()',
                Name.Function),

            include('identifiers'),
        ],

        'whitespace': [
            (r'\n', Text),
            (r'\s+', Text),
            (r'\\\n', Text),
        ],

        'comments': [
            (r'#.*$', Comment),
        ],

        'directives': [
            (r'(@(load-plugin|load-sigs|load|unload))\b.*$', Comment.Preproc),
            (r'(@(DEBUG|DIR|FILENAME|deprecated|if|ifdef|ifndef|else|endif))\b', Comment.Preproc),
            (r'(@prefixes)\s*(\+?=).*$', Comment.Preproc),
        ],

        'attributes': [
            (words(('redef', 'priority', 'log', 'optional', 'default', 'add_func',
                    'delete_func', 'expire_func', 'read_expire', 'write_expire',
                    'create_expire', 'synchronized', 'persistent', 'rotate_interval',
                    'rotate_size', 'encrypt', 'raw_output', 'mergeable', 'error_handler',
                    'broker_allow_complex_type', 'is_assigned', 'is_used',
                    'type_column', 'deprecated', 'on_change', 'backend', 'broker_store'),
                prefix=r'&', suffix=r'\b'),
             Keyword.Pseudo),
        ],

        'types': [
            (words(('any',
                    'enum', 'record', 'set', 'table', 'vector',
                    'function', 'hook', 'event',
                    'addr', 'bool', 'count', 'double', 'file', 'int', 'interval',
                    'pattern', 'port', 'string', 'subnet', 'time'),
                prefix=r'\b', suffix=r'\b'),
             Keyword.Type),

            (r'\b(opaque)(\s+)(of)(\s+)((?:[A-Za-z_][A-Za-z_0-9]*)(?:::(?:[A-Za-z_][A-Za-z_0-9]*))*)\b',
                bygroups(Keyword.Type, Text, Operator.Word, Text, Keyword.Type)),

            (r'\b(type)(\s+)((?:[A-Za-z_][A-Za-z_0-9]*)(?:::(?:[A-Za-z_][A-Za-z_0-9]*))*)(\s*)(:)(\s*)\b(record|enum)\b',
                bygroups(Keyword, Text, Name.Class, Text, Operator, Text, Keyword.Type)),

            (r'\b(type)(\s+)((?:[A-Za-z_][A-Za-z_0-9]*)(?:::(?:[A-Za-z_][A-Za-z_0-9]*))*)(\s*)(:)',
                bygroups(Keyword, Text, Name, Text, Operator)),

            (r'\b(redef)(\s+)(record|enum)(\s+)((?:[A-Za-z_][A-Za-z_0-9]*)(?:::(?:[A-Za-z_][A-Za-z_0-9]*))*)\b',
                bygroups(Keyword, Text, Keyword.Type, Text, Name.Class)),
        ],

        'keywords': [
            (words(('redef', 'export', 'if', 'else', 'for', 'while',
                    'return', 'break', 'next', 'continue', 'fallthrough',
                    'switch', 'default', 'case',
                    'add', 'delete', 'copy',
                    'when', 'timeout', 'schedule'),
                prefix=r'\b', suffix=r'\b'),
             Keyword),
            (r'\b(print)\b', Keyword),
            (r'\b(global|local|const|option)\b', Keyword.Declaration),
            (r'\b(module)(\s+)(([A-Za-z_][A-Za-z_0-9]*)(?:::([A-Za-z_][A-Za-z_0-9]*))*)\b',
                bygroups(Keyword.Namespace, Text, Name.Namespace)),
        ],

        'literals': [
            (r'"', String, 'string'),

            # Not the greatest match for patterns, but generally helps
            # disambiguate between start of a pattern and just a division
            # operator.
            (r'/(?=.*/)', String.Regex, 'regex'),

            (r'\b(T|F)\b', Keyword.Constant),

            # Port
            (r'\b\d{1,5}/(udp|tcp|icmp|unknown)\b', Number),

            # IPv4 Address
            (r'\b(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\b', Number),

            # IPv6 Address (not 100% correct: that takes more effort)
            (r'\[([0-9a-fA-F]{0,4}:){2,7}([0-9a-fA-F]{0,4})?((25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2})\.(25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9]{1,2}))?\]', Number),

            # Numeric
            (r'\b0[xX]' + _hex + r'+\b', Number.Hex),
            (r'\b' + _float + r'\s*(day|hr|min|sec|msec|usec)s?\b', Literal.Date),
            (r'\b' + _float + r'\b', Number.Float),
            (r'\b(\d+)\b', Number.Integer),

            # Hostnames
            (_h + r'(\.' + _h + r')+', String),
        ],

        'operators': [
            (r'[!%*/+<=>~|&^-]', Operator),
            (r'([-+=&|]{2}|[+=!><-]=)', Operator),
            (r'\b(in|as|is|of)\b', Operator.Word),
            (r'\??\$', Operator),
            # Technically, colons are often used for punctuation/sepration.
            # E.g. field name/type separation.
            (r'[?:]', Operator),
        ],

        'punctuation': [
            (r'\?\$', Punctuation),
            (r'[{}()\[\],;:.]', Punctuation),
        ],

        'identifiers': [
            (r'([a-zA-Z_]\w*)(::)', bygroups(Name, Punctuation)),
            (r'[a-zA-Z_]\w*', Name)
        ],

        'string': [
            (r'\\.', String.Escape),
            (r'%-?[0-9]*(\.[0-9]+)?[DTdxsefg]', String.Escape),
            (r'"', String, '#pop'),
            (r'.', String),
        ],

        'regex': [
            (r'\\.', String.Escape),
            (r'/', String.Regex, '#pop'),
            (r'.', String.Regex),
        ],
    }




import unittest
import multiprocessing
import sys
import time
import ipaddress

import broker

class TestCommunication(unittest.TestCase):
    def test_ping(self):
        # --peer-start
        with broker.Endpoint() as ep1, \
             broker.Endpoint() as ep2, \
             ep1.make_subscriber("/test") as s1, \
             ep2.make_subscriber("/test") as s2:

            port = ep1.listen("127.0.0.1", 0)
            ep2.peer("127.0.0.1", port, 1.0)

            # --peer-end

            # --ping-start
            ep2.publish("/test", ["ping"])
            (t, d) = s1.get()
            # t == "/test", d == ["ping"]
            # --ping-end
            self.assertEqual(t, "/test")
            self.assertEqual(d[0], "ping")

            ep1.publish(t, ["pong"])

            while True:
                # This loop exists just for sake of test coverage for "poll()"
                msgs = s2.poll()

                if msgs:
                    self.assertEqual(len(msgs), 1)
                    (t, d) = msgs[0]
                    break;

                time.sleep(0.1)

            self.assertEqual(t, "/test")
            self.assertEqual(d[0], "pong")

    def test_messages(self):
        with broker.Endpoint() as ep1, \
             broker.Endpoint() as ep2, \
             ep1.make_subscriber("/test") as s1:

            port = ep1.listen("127.0.0.1", 0)
            ep2.peer("127.0.0.1", port, 1.0)

            msg0 = ("/test/1", ())
            ep2.publish(*msg0)

            # --messages-start
            msg1 = ("/test/2", (1, 2, 3))
            msg2 = ("/test/3", (42, "foo", {"a": "A", "b": ipaddress.IPv4Address('1.2.3.4')}))
            ep2.publish_batch(msg1, msg2)
            # --messages-end

            msgs = s1.get(3)
            self.assertFalse(s1.available())

            self.assertEqual(msgs[0], msg0)
            self.assertEqual(msgs[1], msg1)
            self.assertEqual(msgs[2], msg2)

            # These results are not (all) immutable: try modifying the third
            # value (the dict) of the last message above.
            dict_data = msgs[2][1][2]
            self.assertEqual(len(dict_data), 2)
            dict_data["c"] = "not immutable"
            self.assertEqual(len(dict_data), 3)

    def test_immutable_messages(self):
        with broker.Endpoint() as ep1, \
             broker.Endpoint() as ep2, \
             ep1.make_safe_subscriber("/test") as s1:

            port = ep1.listen("127.0.0.1", 0)
            ep2.peer("127.0.0.1", port, 1.0)

            msg = ("/test/1", ({"a": "A"}, set([1,2,3]), ('a', 'b', 'c')))
            ep2.publish(*msg)

            topic, (dict_data, set_data, tuple_data) = s1.get()

            # The return values are immutable, so each of the following triggers
            # a type-specific exception.
            with self.assertRaises(TypeError):
                # 'mappingproxy' object does not support item assignment
                dict_data["b"] = "B"
            with self.assertRaises(AttributeError):
                # 'frozenset' object has no attribute 'add'
                set_data.add(4)
            with self.assertRaises(TypeError):
                # 'tuple' object does not support item assignment
                tuple_data[3] = 'd'

    def test_publisher(self):
        with broker.Endpoint() as ep1, \
             broker.Endpoint() as ep2, \
             ep1.make_subscriber("/test") as s1, \
             ep2.make_publisher("/test") as p2:

            port = ep1.listen("127.0.0.1", 0)
            ep2.peer("127.0.0.1", port, 1.0)

            p2.publish([1, 2, 3])
            p2.publish_batch(["a", "b", "c"], [True, False])

            msgs = s1.get(3)
            self.assertFalse(s1.available())

            self.assertEqual(msgs[0], ("/test", (1, 2, 3)))
            self.assertEqual(msgs[1], ("/test", ("a", "b", "c")))
            self.assertEqual(msgs[2], ("/test", (True, False)))

    def test_status_subscriber(self):
        # --status-start
        with broker.Endpoint() as ep1, \
             broker.Endpoint() as ep2, \
             ep1.make_status_subscriber(True) as es1, \
             ep2.make_status_subscriber(True) as es2:

            port = ep1.listen("127.0.0.1", 0)
            ep2.peer("127.0.0.1", port, 1.0)
            st1 = es1.get()
            st2 = es2.get()
            # st1.code() == broker.SC.PeerAdded, st2.code() == broker.SC.PeerAdded
            # --status-end

            self.assertEqual(st1.code(), broker.SC.PeerAdded)
            self.assertEqual(st1.context().network.get().address, "127.0.0.1")
            self.assertEqual(st2.code(), broker.SC.PeerAdded)
            self.assertEqual(st2.context().network.get().address, "127.0.0.1")

    def test_status_subscriber_error(self):
        # --error-start
        with broker.Endpoint() as ep1, \
             ep1.make_status_subscriber() as es1:

            r = ep1.peer("127.0.0.1", 1947, 0.0) # Try unavailable port, no retry
            self.assertEqual(r, False) # Not shown in docs.
            st1 = es1.get()
            # s1.code() == broker.EC.PeerUnavailable
            # --error-end
            self.assertEqual(st1.code(), broker.EC.PeerUnavailable)

            # Async version.
            ep1.peer_nosync("127.0.0.1", 1947, 1.0)
            st1 = es1.get()
            self.assertEqual(st1.code(), broker.EC.PeerUnavailable)

            st1 = es1.get()
            self.assertEqual(st1.code(), broker.EC.PeerUnavailable)

    def test_idle_endpoint(self):
        with broker.Endpoint() as ep1, \
             ep1.make_status_subscriber() as es1, \
             ep1.make_subscriber("/test") as s1:

            pass

if __name__ == '__main__':
    unittest.main(verbosity=3)



# Zeekctl test plugin that uses plugin state variables.

import ZeekControl.plugin

class StateTest(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(StateTest, self).__init__(apiversion=1)

    def name(self):
        return "statetest"

    def pluginVersion(self):
        return 1

    def init(self):
        sv = self.getState("statevar")
        if not sv:
            sv = 0

        self.setState("statevar", sv + 1)

        with open("state.out", "w") as f:
            f.write("statevar: %s\n" % self.getState("statevar"))

        return True



#!/usr/bin/env python3

# Acld interface for the Network Control Framework of Zeek, using Broker.

from __future__ import print_function

import argparse
import errno
import fcntl, os
import logging
import random
import re
import socket
import string
import sys
import ipaddress
import datetime
import time

import broker
import broker.zeek
from select import select
from logging.handlers import TimedRotatingFileHandler

MAX16INT = 2**16 - 1

def parseArgs():
    defaultuser = os.getlogin()
    defaulthost = socket.gethostname()
    defaultacldhost = '127.0.0.1'

    parser = argparse.ArgumentParser()
    parser.add_argument('--listen', default="127.0.0.1",
        help="Address to listen on for connections (default: %(default)s)")
    parser.add_argument('--port', type=int, default=9999,
        help="Port to listen on for connections (default: %(default)s)")
    parser.add_argument('--acld_host', metavar='HOST', action='append',
        help='ACLD hosts to connect to (default: %s)' % defaultacldhost)
    parser.add_argument('--acld_port', metavar='PORT', type=int, default=11775,
        help="ACLD port to connect to (default: %(default)s)")
    parser.add_argument('--log-user', default=defaultuser,
        help='user name provided to acld (default: %(default)s)')
    parser.add_argument('--log-host', default=defaulthost,
        help='host name provided to acld (default: %(default)s)')
    parser.add_argument('--topic', default="zeek/event/pacf",
        help="Topic to subscribe to. (default: %(default)s)")
    parser.add_argument('--debug', const=logging.DEBUG, action='store_const',
        default=logging.INFO,
        help="Enable debug output")
    parser.add_argument('--logfile',
        help="Filename of logfile. If not given, logs to stdout")
    parser.add_argument('--rotate', action="store_true",
        help="If logging to file and --rotate is specified, log will rotate at midnight")

    args = parser.parse_args()
    if not args.acld_host:
        args.acld_host = [defaultacldhost]
    return args

def hostportpair(host, port):
    """Host is an ip address or ip address and port,
       port is the default port.
       return a host-port pair"""
    tup = host.split(',', 1)
    if len(tup) == 2:
        host = tup[0]
        sport = tup[1]
        if not sport.isdigit():
            self.logger.error('%s: port must be numeric' % host)
            sys.exit(-1)
        port = int(sport)
    if port <= 0 or port > MAX16INT:
        self.logger.error('%s: port must be > 0 and < %d ' % (host, MAX16INT))
        sys.exit(-1)
    return host, port

class Listen(object):
    TIMEOUT_INITIAL = 0.25
    TIMEOUT_MAX = 8.0

    def __init__(self, queue, host, port, acld_hosts, acld_port, log_user, log_host):
        self.logger = logging.getLogger("brokerlisten")

        self.queuename = queue
        self.epl = broker.Endpoint()
        self.epl.listen(host, port)
        self.status_subscriber = self.epl.make_status_subscriber(True)
        self.subscriber = self.epl.make_subscriber(self.queuename)

        # Create a random list of host port pairs
        self.acld_hosts = []
        for host in acld_hosts:
            self.acld_hosts.append(hostportpair(host, acld_port))
        random.shuffle(self.acld_hosts)

        self.ident = '{%s@%s}' % (log_user, log_host)
        self.remote_ident = '?'

        self.sock = None

        self.waiting = {}
        self.buffer = ''

        self.acldstring = False
        self.acldcmd = {}

        # try to connect to acld
        self.connect()

    def connect(self):
        """Round robin across multiple aclds with exponential backoff"""
        if self.sock:
            self.sock.close()
            self.sock = None;
        # No delay on first retry with multiple aclds
        if len(self.acld_hosts) > 1:
            timeout = 0.0
        else:
            timeout = self.TIMEOUT_INITIAL
        while True:
            self.sock = socket.socket()
            hostpair = self.get_hostpair()
            self.remote_ident = '[%s].%d' % (hostpair[0], hostpair[1])
            self.logger.debug('%s Connecting' % self.remote_ident)
            try:
                self.sock.connect(hostpair)
            except socket.error as e:
                self.logger.error('%s %s' % (self.remote_ident, e.strerror))
                time.sleep(timeout)
                if not timeout:
                    timeout = self.TIMEOUT_INITIAL
                else:
                    timeout *= 2
                    if timeout > self.TIMEOUT_MAX:
                        timeout = self.TIMEOUT_MAX
                continue

            fcntl.fcntl(self.sock, fcntl.F_SETFL, os.O_NONBLOCK)
            self.logger.info('%s Connected' % self.remote_ident)
            break

    def get_hostpair(self):
        """Round robin multiple ACLD hosts"""
        hostport = self.acld_hosts.pop(0)
        self.acld_hosts.append(hostport)
        return hostport

    def listen_loop(self):
        self.logger.debug("Broker loop...")

        while 1==1:
            self.logger.debug("Waiting for broker message")
            readable, writable, exceptional = select(
                    [self.status_subscriber.fd(),
                     self.subscriber.fd(), self.sock],
                    [], [])

            if ( self.status_subscriber.fd() in readable ):
                self.logger.debug("Got broker status message")
                msg = self.status_subscriber.get()
                self._handle_broker_message(msg)
            elif ( self.subscriber.fd() in readable ):
                self.logger.debug("Got broker message")
                msg = self.subscriber.get()
                self._handle_broker_message(msg)
            elif ( self.sock in readable ):
                line = self.read_acld()
                while line != None:
                    self.logger.info("Received from ACLD: %s", line)
                    self.parse_acld(line)
                    line = self.read_acld()
                continue


    def parse_acld(self, line):
        line = line.rstrip("\r")
        if self.acldstring == False:
            items = line.split(" ")
            if len(items) == 3:
                ts, cookie, command = items
                more = None
            elif len(items) == 4:
                ts, cookie, command, more = items
            else:
                self.logger.error("Could not parse acld line: %s", line)
                return

            self.acldcmd = {'ts': ts, 'cookie': cookie, 'command': command}
            if more != None:
                self.acldstring = True
                self.acldcmd['comment'] = ""
                if more != "-":
                    self.logger.error("Parse error while parsing acld line: %s?", more)
            else:
                self.execute_acld()
        else:
            if line == ".":
                self.acldstring = False
                self.execute_acld()
            else:
                self.acldcmd['comment']+=line

    def execute_acld(self):
        cmd = self.acldcmd['command']
        cookie = int(self.acldcmd['cookie'])
        comment = self.acldcmd.get('comment', "")

        if cmd == "acld":
            # we get this when connecting
            self.logger.info('%s acld connection succesful' % self.remote_ident)
            return

        if cookie in self.waiting:
            msg = self.waiting[cookie]
            del self.waiting[cookie]

            if "-failed" in cmd:
                if re.search(".* is on the whitelist .*", comment):
                    self.rule_event("exists", msg['id'], msg['arule'], msg['rule'], comment)
                else:
                    self.rule_event("error", msg['id'], msg['arule'], msg['rule'], comment)
            elif re.search("Note: .* is already ", comment):
                self.rule_event("exists", msg['id'], msg['arule'], msg['rule'], comment)
            else:
                type = "added"
                if msg['add'] == False:
                    type = "removed"
                self.rule_event(type, msg['id'], msg['arule'], msg['rule'], comment)

        else:
            self.logger.warning("Got response to cookie %d we did not send. Ignoring", cookie)
            return

    def read_acld(self):
        try:
            data = self.sock.recv(4096)
            if len(data) == 0:
                self.logger.warning('%s Disconnected' % self.remote_ident)
                self.connect()
            self.buffer += data.decode("utf-8")
        except socket.error as e:
            err = e.args[0]
            if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                # socket not ready yet, just continue and see if something
                # is still in the buffer
                pass
            else:
                self.logger.error(e)
                sys.exit(-1)
                return

        if self.buffer.find("\r\n") != -1:
            line, self.buffer = self.buffer.split("\r\n", 1)
            return line
        else:
            return None

    def _handle_broker_message(self, m):
        if isinstance(m, broker.Status):
            if m.code() == broker.SC.PeerAdded:
                self.logger.info("Incoming connection established.")
                return

            return

        if type(m).__name__ != "tuple":
            self.logger.error("Unexpected type %s, expected tuple.", type(m).__name__)
            return

        if len(m) < 1:
            self.logger.error("Tuple without content?")
            return

        (topic, event) = m
        ev = broker.zeek.Event(event)
        event_name = ev.name()

        if event_name == "NetControl::acld_add_rule":
            self.add_remove_rule(event_name, ev.args(), True)
        elif event_name == "NetControl::acld_remove_rule":
            self.add_remove_rule(event_name, ev.args(), False)
        elif event_name == "NetControl::acld_rule_added":
            pass
        elif event_name == "NetControl::acld_rule_removed":
            pass
        elif event_name == "NetControl::acld_rule_error":
            pass
        elif event_name == "NetControl::acld_rule_exists":
            pass
        else:
            self.logger.error("Unknown event %s", event_name)
            return

    def add_remove_rule(self, name, m, add):
        print(m)
        if ( len(m) != 3 ) or ( not isinstance(m[0], broker.Count) ) or (not isinstance(m[1], tuple) ) or ( not isinstance(m[2], tuple) ):
            self.logger.error("wrong number of elements or type in tuple for acld_add|remove_rule")
            return

        id = m[0].value
        arule = self.record_to_record("acldrule", m[2])

        self.logger.info("Got event %s. id=%d, arule: %s", name, id, arule)

        cmd = arule['command'] + " " + str(arule['cookie']) + " " + arule['arg'] + " -"
        sendlist = [cmd, self.ident]
        if 'comment' in arule and arule['comment'] != None and len(arule['comment']) > 0:
            sendlist.append(arule['comment'])
        sendlist.append(".")

        self.waiting[arule['cookie']] = {'add': add, 'cmd': cmd, 'id': m[0], 'rule': m[1], 'arule': m[2]}
        self.logger.info("Sending to ACLD: %s", ", ".join(sendlist))
        self.sock.sendall(("\r\n".join(sendlist)+"\r\n").encode())

    def rule_event(self, event, id, arule, rule, msg):
        arule = self.record_to_record("acldrule", arule)
        self.logger.info("Sending to Zeek: NetControl::acld_rule_%s id=%d, arule=%s, msg=%s", event, id.value, arule, msg)

        ev = broker.zeek.Event("NetControl::acld_rule_"+event, id, rule, msg)
        self.epl.publish(self.queuename, ev)

    def record_to_record(self, name, m):
        if not isinstance(m, tuple):
            self.logger.error("Got non record element")

        rec = m

        elements = None
        if name == "acldrule":
            elements = ['command', 'cookie', 'arg', 'comment']
        else:
            self.logger.error("Unknown record type %s", name)
            return

        dict = {}
        for i in range(0, len(elements)):
            if rec[i] is None:
                dict[elements[i]] = None
                continue
            elif isinstance(rec[i], tuple):
                dict[elements[i]] = self.record_to_record(name+"->"+elements[i], rec[i])
                continue

            dict[elements[i]] = self.convert_element(rec[i])

        return dict

    def convert_element(self, el):
        if isinstance(el, broker.Count):
            return el.value

        if isinstance(el, ipaddress.IPv4Address):
            return str(el);

        if isinstance(el, ipaddress.IPv6Address):
            return str(el);

        if isinstance(el, ipaddress.IPv4Network):
            return str(el);

        if isinstance(el, ipaddress.IPv6Network):
            return str(el);

        if isinstance(el, broker.Port):
            p = str(el)
            ex = re.compile('([0-9]+)(.*)')
            res = ex.match(p)
            return (res.group(1), res.group(2))

        if isinstance(el, broker.Enum):
            tmp = el.name
            return re.sub(r'.*::', r'', tmp)

        if isinstance(el, list):
            return [convertElement(ell) for ell in el];

        if isinstance(el, datetime.datetime):
            return el

        if isinstance(el, datetime.timedelta):
            return el

        if isinstance(el, int):
            return el

        if isinstance(el, str):
            return el

        logger.error("Unsupported type %s", type(el) )
        return el;

args = parseArgs()
logger = logging.getLogger('')
logger.setLevel(args.debug)

handler = None

if args.logfile:
    if args.rotate:
        handler = TimedRotatingFileHandler(args.logfile, 'midnight')
    else:
        handler = logging.FileHandler(args.logfile);
else:
    handler = logging.StreamHandler(sys.stdout)

formatter = logging.Formatter('%(created).6f:%(name)s:%(levelname)s:%(message)s')
handler.setFormatter(formatter)

logger.addHandler(handler)

logging.info("Starting acld.py...")
brocon = Listen(args.topic, args.listen, args.port, args.acld_host,
    args.acld_port, args.log_user, args.log_host)
try:
    brocon.listen_loop()
except KeyboardInterrupt:
    pass



import sphinx
import re
from docutils import nodes

# This extension adds a 'literal-emph' directive that operates the same
# as the 'code-block' directive except that it additionally understands
# the **strong emphasis** markup, allowing custom rendering of it to be
# substituted in the final literal block (e.g. HTML adds <strong> elements).
# Adding " (no-emph)" to the end of a line within the 'literal-emph' content
# disables substitutions for that line.

class LiteralEmphNode(nodes.General, nodes.Element):
    pass

class LiteralEmph(sphinx.directives.code.CodeBlock):
    def run(self):
        node = LiteralEmphNode()
        node += super().run()
        return [node]

def visit_litemph_node(self, node):
    pass

def depart_litemph_node(self, node):
    text = self.body[-1]
    text = re.sub(r'\*\*(.*?)\*\*(?!.* \(no-emph\)\n)',
                  r'<strong>\1</strong>',
                  text)
    text = re.sub(r'(.*) \(no-emph\)\n', r'\1\n', text)
    self.body[-1] = text

def setup(app):
    app.add_directive("literal-emph", LiteralEmph)
    app.add_node(LiteralEmphNode,
                 html=(visit_litemph_node, depart_litemph_node))



import sys, socket

# Keep track of the test case number so that any error message is more useful.
testnum = 0

def testcase(success, errmsg):
    global testnum
    testnum += 1

    if not success:
        sys.stderr.write("Test #%d: %s\n" % (testnum, errmsg))
        sys.exit(1)

def ipv4binary(ipstr):
    return socket.inet_pton(socket.AF_INET, ipstr)

def ipv6binary(ipstr):
    return socket.inet_pton(socket.AF_INET6, ipstr)




import os, sys, zipfile, subprocess, shutil

from subprocess import PIPE

# -- setup and cleanup --------------------------------------------------------

test_recording = 'zeek-dns-traffic-recording'

class Environment:
    def __init__(self, executable_path):
        self.exe = executable_path
        self.test_dir = os.environ.get('BROKER_TEST_DIR')
        if not os.path.isdir(self.test_dir):
            raise RuntimeError('environment variable BROKER_TEST_DIR is not a valid directory')
        self.input_dir = os.path.join(self.test_dir, '.tmp', 'broker-cluster-benchmark')
        self.recording_dir = os.path.join(self.input_dir, test_recording)

    def __enter__(self):
        self.clear_environment()
        self.prepare_environment()
        return self

    def __exit__(self, type, value, tb):
        self.clear_environment()

    def prepare_environment(self):
        os.makedirs(self.input_dir)
        file_name = test_recording + '.zip'
        file_path = os.path.join(self.test_dir, 'integration', file_name)
        with zipfile.ZipFile(file_path, 'r') as zip_ref:
            zip_ref.extractall(self.input_dir)
        with open(os.path.join(self.recording_dir, 'expected-tpl.conf'), mode='r') as f:
            format_vars = {'path': self.recording_dir}
            self.expected = f.read() % format_vars

    def clear_environment(self):
        if os.path.isdir(self.input_dir):
            shutil.rmtree(self.input_dir)

# -- integration testing ------------------------------------------------------

def test_config_generation(exe, recording_dir, expected):
    dirs = [f.path for f in os.scandir(recording_dir) if f.is_dir()]
    cmd = [exe, '--caf.logger.console.verbosity=quiet', '--mode=generate-config'] + dirs
    with subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE, close_fds=True, universal_newlines=True) as proc:
        output, errors = proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError('failed to generate config: ' + errors)
        if output != expected:
            sys.stderr.write('*** ERROR: generate-config procuded wrong result\n')
            sys.stderr.write('\n*** EXPECTED:\n')
            sys.stderr.write(expected)
            sys.stderr.write('\n*** GOT:\n')
            sys.stderr.write(output)
            sys.exit(1)
        print('test_config_generation: pass')
        return output

def run_benchmark(exe, config):
    cmd = [exe, '--caf.logger.console.verbosity=quiet', '--cluster-config-file=-']
    with subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, close_fds=True, universal_newlines=True) as proc:
        proc.stdin.write(config)
        proc.stdin.close()
        proc.stdin = None
        output, errors = proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError('failed to run the benchmark: ' + errors)
        print('run_benchmark: pass')
        sys.stdout.write('*** Benchmark output:\n')
        sys.stdout.write(output)


# -- main ---------------------------------------------------------------------

if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise RuntimeError('expected exactly one argument: path to broker-cluster-benchmark')
    with Environment(sys.argv[1]) as e:
        config = test_config_generation(e.exe, e.recording_dir, e.expected)
        run_benchmark(e.exe, config)




import unittest
import multiprocessing
import sys
import time

import broker

def cleanup(es, ss):
    for s in ss:
        if s:
            s.reset()

    for e in es:
        e.shutdown()

def setup_peers(opts1=None, opts2=None, opts3=None, opts4=None, create_s1=True, create_s2=True, create_s3=True, create_s4=True):
    def cfg(opts):
        return broker.Configuration(opts) if opts else broker.Configuration(broker.BrokerOptions())
    ep1 = broker.Endpoint(cfg(opts1))
    ep2 = broker.Endpoint(cfg(opts2))
    ep3 = broker.Endpoint(cfg(opts3))
    ep4 = broker.Endpoint(cfg(opts4))

    s1 = ep1.make_subscriber("/test/") if create_s1 else None
    s2 = ep2.make_subscriber("/test/") if create_s2 else None
    s3 = ep3.make_subscriber("/test/") if create_s3 else None
    s4 = ep4.make_subscriber("/test/") if create_s4 else None

    p2 = ep2.listen("127.0.0.1", 0)
    p3 = ep3.listen("127.0.0.1", 0)
    p4 = ep4.listen("127.0.0.1", 0)

    # ep1 <-> ep2 <-> ep3 <-> ep4
    ep1.peer("127.0.0.1", p2, 1.0)
    ep2.peer("127.0.0.1", p3, 1.0)
    ep3.peer("127.0.0.1", p4, 1.0)

    return ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))

class TestCommunication(unittest.TestCase):
    def test_two_hops(self):
        # Two hops that are subscribed, so they'll forward.
        ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4)) = setup_peers()

        ep1.publish("/test/foo", "Foo!")
        ep4.publish("/test/bar", "Bar!")

        x = s4.get()
        self.assertEqual(x, ('/test/foo', 'Foo!'))
        x = s1.get()
        self.assertEqual(x, ('/test/bar', 'Bar!'))
        cleanup((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))

    def test_two_hops_with_forward(self):
        # Two hops that are not subscribed, but configured to forward.
        ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4)) = setup_peers(create_s2=False, create_s3=False)

        ep2.forward("/test/");
        ep3.forward("/test/");
        time.sleep(1) # give time to take effect.

        ep1.publish("/test/foo", "Foo!")
        ep4.publish("/test/bar", "Bar!")

        x = s4.get()
        self.assertEqual(x, ('/test/foo', 'Foo!'))
        x = s1.get()
        self.assertEqual(x, ('/test/bar', 'Bar!'))
        cleanup((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))

    def test_two_hops_forwarding_disabled(self):
        # Two hops that are subscribed, so they would forward but we disable.
        no_forward = broker.BrokerOptions()
        no_forward.forward = False

        ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4)) = setup_peers(opts2=no_forward)

        ep1.publish("/test/foo", "Foo!") # Shouldn't arrive
        x = s4.get(1.0)
        self.assertEqual(x, None)
        cleanup((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))

    def test_two_hops_without_forward(self):
        # Two hops that are not subscribed, and hence don't forward.
        ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4)) = setup_peers(create_s2=False, create_s3=False)

        ep1.publish("/test/foo", "Foo!")
        x = s4.get(1.0)
        cleanup((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))
        self.assertEqual(x, None)

    def test_two_hops_ttl(self):
        ttl1 = broker.BrokerOptions()
        ttl1.ttl = 2
        ((ep1, ep2, ep3, ep4), (s1, s2, s3, s4)) = setup_peers(opts1=ttl1)

        ep1.publish("/test/foo", "Foo!")

        x = s2.get(1.0)
        self.assertEqual(x, ('/test/foo', 'Foo!'))
        x = s3.get(1.0)
        self.assertEqual(x, ('/test/foo', 'Foo!'))
        x = s4.get(1.0)
        self.assertEqual(x, None) # Doesn't get here anymore.
        cleanup((ep1, ep2, ep3, ep4), (s1, s2, s3, s4))

if __name__ == '__main__':
    #TestCommunication().test_two_hops()
    unittest.main(verbosity=3)
